/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * Copyright (C) 1986, Sun Microsystems, Inc.
 * Patched by Olaf Kirch okir@monad.swb.de to make it more secure, 
 * ANSIfied it, and some cosmetic changes.
 */

/*
 * Keyserver 
 * Store secret keys per uid. Do public key encryption and decryption
 * operations. Generate "random" keys. Do not talk to anything but a local root
 * process (by checking that the source port < IPPORT_RESERVED and by binding
 * to the loopback address). 
 */

#include <sys/file.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpc/pmap_clnt.h>
#include <pwd.h>
#include <des_crypt.h>
#include <stdio.h>

#include "keyserv.h"

static int	getrootkey(des_block *master, int prompt);
static void	keyprogram(struct svc_req *rqstp, SVCXPRT *transp);
static void	usage(void);

static SVCXPRT	*udptransp,
		*tcptransp;
static des_block masterkey;

int
main(int argc, char **argv)
{
	int	c;
	int	nflag = 0;
	int	foreground = 0;

	xlog_open("keyserv");

	while ((c = getopt(argc, argv, "nd:")) != EOF) {
		switch (c) {
		case 'd':
			xlog_sconfig(optarg, 1);
			foreground = 1;
			break;
		case 'n':
			nflag = 1;
			break;
		default:
			usage();
		}
	}
	if (argc != optind)
		usage();

	/*
	 * Initialize 
	 */
	umask(066);	/* paranoia */
	if (geteuid() != 0)
		xlog(L_FATAL, "must be run as root\n");

	setmodulus(HEXMODULUS);
	openstore();
	getrootkey(&masterkey, nflag);
	readkeys();

	/*
	 * create the service, register it, and run 
	 */
	if ((udptransp = svcudp_create(RPC_ANYSOCK)) == NULL)
		xlog(L_FATAL, "unable to create udp service\n");
	if ((tcptransp = svctcp_create(RPC_ANYSOCK, 2048, 2048)) == NULL)
		xlog(L_FATAL, "unable to create tcp service\n");

	pmap_unset(KEY_PROG, KEY_VERS);
	if (!svc_register(udptransp, KEY_PROG, KEY_VERS, keyprogram,
			  IPPROTO_UDP))
		xlog(L_FATAL, "unable to register udp service\n");
	if (!svc_register(tcptransp, KEY_PROG, KEY_VERS, keyprogram,
			  IPPROTO_TCP))
		xlog(L_FATAL, "unable to register tcp service\n");

	if (!foreground) {
		if ((c = fork()) < 0)
			xlog(L_FATAL, "cannot fork\n");
		if (c != 0)
			exit(0);
		setsid();
		xlog_background();
	}

	svc_run();

	xlog(L_FATAL, "svc_run returned!\n");
	return(1);
}

static void
usage(void)
{
	xlog(L_FATAL, "usage: keyserv [-nd]\n");
}

/*
 * In the event that we don't get a root password, we try to randomize the
 * master key the best we can.
 * FIXME: This is not good enough; boot time is too easy to guess.
 */
#ifndef MUST_USE_DEVRANDOM
static void
oldrandomize(des_block *master)
{
	struct timeval tv;
	int	i;
	int	seed;
	int	shift;

	seed = 0;
	for (i = 0; i < 1024; i++) {
		gettimeofday(&tv, (struct timezone *) NULL);
		shift = i % 8 * sizeof(int);
		seed ^= (tv.tv_usec << shift) | (tv.tv_usec >> (32 - shift));
	}
	srandom(seed);
	master->key.low = random();
	master->key.high = random();
	srandom(seed);
}
#endif

static void
randomize(des_block *master)
{
	int	fd, got, n;
	char	*buf = (char *) &master->key;

	if ((fd = open("/dev/random", O_RDONLY)) < 0) {
		xlog(L_ERROR, "can't create root key from /dev/random: %s",
						strerror(errno));
#ifndef MUST_USE_DEVRANDOM
		xlog(L_ERROR, "resorting to old randomize() function");
		oldrandomize(master);
#else
		xlog(L_FATAL, "keyserver aborting");
#endif
	}

	for (got = 0; got < 8; got += n) {
		if ((n = read(fd, buf+got, 8-got)) <= 0) {
			xlog(L_FATAL, "failed to read from /dev/random: %s",
					strerror(errno));
		}
	}
	close(fd);
}

/*
 * Try to get root's secret key, by prompting if terminal is a tty, else trying
 * from standard input. 
 */
static int
getrootkey(des_block *master, int prompt)
{
	char *passwd;
	char name[MAXNETNAMELEN + 1];
	char secret[HEXKEYBYTES + 1];
	int fd;

	if (!prompt) {
		/*
		 * Read secret key out of $ROOTKEY 
		 */
		if ((fd = open(ROOTKEY_FILE, O_RDONLY, 0)) < 0) {
			randomize(master);
			return (0);
		}
		if (read(fd, secret, HEXKEYBYTES) < 0) {
			xlog(L_ERROR, "invalid %s\n", ROOTKEY_FILE);
			close(fd);
			return (0);
		}
		close(fd);
		secret[HEXKEYBYTES] = 0;
	} else {
		/*
		 * Decrypt yellow pages entry to get secret key 
		 */
		passwd = getpass("root password:");
		passwd2des(passwd, master->c);
		getnetname(name);
		if (!getsecretkey(name, secret, passwd)) {
			xlog(L_ERROR, "can't find %s's secret key\n", name);
			return (0);
		}
		if (secret[0] == 0) {
			xlog(L_ERROR, "invalid password for %s\n", name);
			return (0);
		}
	}
	pk_setkey(0, secret);
	return (1);
}


/*
 * Procedures to implement RPC service
 */

static char *
strstatus(keystatus status)
{
	switch (status) {
	case KEY_SUCCESS:
		return ("KEY_SUCCESS");
	case KEY_NOSECRET:
		return ("KEY_NOSECRET");
	case KEY_UNKNOWN:
		return ("KEY_UNKNOWN");
	case KEY_SYSTEMERR:
		return ("KEY_SYSTEMERR");
	}
	return ("(bad result code)");
}

static keystatus *
key_set_1(short uid, keybuf key)
{
	static keystatus status;

	status = pk_setkey(uid, key);
	xlog(D_CALL, "set(%d, %.*s) = %s\n", uid, sizeof(keybuf), key,
						strstatus(status));
	return (&status);
}



static cryptkeyres *
key_encrypt_1(short uid, cryptkeyarg *arg)
{
	static cryptkeyres res;

	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_encrypt(uid, arg->remotename,
				&res.cryptkeyres_u.deskey);

	if (res.status == KEY_SUCCESS) {
		xlog(D_CALL, "encrypt(%d, %s, %08x%08x) = %08lx%08lx\n",
				uid, arg->remotename,
				arg->deskey.key.high, arg->deskey.key.low,
				res.cryptkeyres_u.deskey.key.high,
				res.cryptkeyres_u.deskey.key.low);
	} else {
		xlog(D_CALL, "encrypt(%d, %s, %08x%08x) = %s\n",
				uid, arg->remotename,
				arg->deskey.key.high, arg->deskey.key.low,
				strstatus(res.status));
	}
	return (&res);
}

static cryptkeyres *
key_decrypt_1(short uid, cryptkeyarg *arg)
{
	static cryptkeyres res;

	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_decrypt(uid, arg->remotename,
				&res.cryptkeyres_u.deskey);

	if (res.status == KEY_SUCCESS) {
		xlog(D_CALL, "decrypt(%d, %s, %08x%08x) = %08lx%08lx\n",
				uid, arg->remotename,
				arg->deskey.key.high, arg->deskey.key.low,
				res.cryptkeyres_u.deskey.key.high,
				res.cryptkeyres_u.deskey.key.low);
	} else {
		xlog(D_CALL, "decrypt(%d, %s, %08x%08x) = %s\n",
				uid, arg->remotename,
				arg->deskey.key.high, arg->deskey.key.low,
				strstatus(res.status));
	}

	return (&res);
}

static des_block *
key_gen_1(void)
{
	struct timeval time;
	static des_block keygen;
	static des_block key;

	gettimeofday(&time, (struct timezone *) NULL);
	keygen.key.high += (time.tv_sec ^ time.tv_usec);
	keygen.key.low += (time.tv_sec ^ time.tv_usec);
	ecb_crypt(masterkey.c, keygen.c, sizeof(keygen), DES_ENCRYPT | DES_HW);
	key = keygen;
	des_setparity(key.c);

	xlog(D_CALL, "gen() = %08x%08x\n", key.key.high, key.key.low);

	return (&key);
}

static getcredres *
key_getcred_1(short uid, netnamestr *name)
{
	static getcredres res;
	static int	gids[NGROUPS];
	struct unixcred	*cred;

	cred = &res.getcredres_u.cred;
	cred->gids.gids_val = gids;
	if (!netname2user(*name, &cred->uid, &cred->gid,
			  &cred->gids.gids_len, gids)) {
		res.status = KEY_UNKNOWN;
	} else {
		res.status = KEY_SUCCESS;
	}

	if (res.status == KEY_SUCCESS) {
		xlog(D_CALL, "getcred(%s) = uid=%d,gid=%d,grouplen=%d\n",
			*name, cred->uid, cred->gid, cred->gids.gids_len);
	} else {
		xlog(D_CALL, "getcred(%s) = %s\n",
			*name, strstatus(res.status));
	}

	return (&res);
}


/*
 * RPC boilerplate 
 */
static void
keyprogram(struct svc_req *rqstp, SVCXPRT *transp)
{
	union {
		keybuf key_set_1_arg;
		cryptkeyarg key_encrypt_1_arg;
		cryptkeyarg key_decrypt_1_arg;
		des_block key_gen_1_arg;
	} argument;
	char	*result;

	xdrproc_t	xdr_argument, xdr_result;
	char		*(*local) ();
	struct sockaddr_in remote;
	int		uid;
	int		check_auth;
	char		*procname;

	switch (rqstp->rq_proc) {
	case NULLPROC:
		svc_sendreply(transp, (xdrproc_t)xdr_void, (char *) NULL);
		return;

	case KEY_SET:
		procname = "key_set";
		xdr_argument = (xdrproc_t) xdr_keybuf;
		xdr_result = (xdrproc_t) xdr_int;
		local = (char *(*)()) key_set_1;
		check_auth = 1;
		break;

	case KEY_ENCRYPT:
		procname = "key_encrypt";
		xdr_argument = (xdrproc_t) xdr_cryptkeyarg;
		xdr_result = (xdrproc_t) xdr_cryptkeyres;
		local = (char *(*)()) key_encrypt_1;
		check_auth = 1;
		break;

	case KEY_DECRYPT:
		procname = "key_decrypt";
		xdr_argument = (xdrproc_t) xdr_cryptkeyarg;
		xdr_result = (xdrproc_t) xdr_cryptkeyres;
		local = (char *(*)()) key_decrypt_1;
		check_auth = 1;
		break;

	case KEY_GEN:
		procname = "key_gen";
		xdr_argument = (xdrproc_t) xdr_void;
		xdr_result = (xdrproc_t) xdr_des_block;
		local = (char *(*)()) key_gen_1;
		check_auth = 0;
		break;

	case KEY_GETCRED:
		procname = "key_getcred";
		xdr_argument = (xdrproc_t) xdr_netnamestr;
		xdr_result = (xdrproc_t) xdr_getcredres;
		local = (char *(*)()) key_getcred_1;
		check_auth = 0;
		break;

	default:
		svcerr_noproc(transp);
		return;
	}
	if (check_auth) {
		remote = *svc_getcaller(transp);

		if (ntohs(remote.sin_port) >= IPPORT_RESERVED ||
		    ntohs(remote.sin_port) < IPPORT_RESERVED/2 ||
		    ntohl(remote.sin_addr.s_addr) != INADDR_LOOPBACK) {
			xlog(L_WARNING, "procedure %s called from %s:%d\n",
				procname, inet_ntoa(remote.sin_addr),
				ntohs(remote.sin_port));
			svcerr_weakauth(transp);
			return;
		}
#if 0
		if (transp != tcptransp) {
			xlog(L_WARNING, "procedure %s called on UDP socket!\n",
						procname);
			svcerr_weakauth(transp);
			return;
		}
#endif
		if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
			xlog(L_WARNING,
				"procedure %s called without AUTH_UNIX\n",
				procname);
			svcerr_weakauth(transp);
			return;
		}
		uid = ((struct authunix_parms *) rqstp->rq_clntcred)->aup_uid;
	}
	bzero((char *) &argument, sizeof(argument));
	if (!svc_getargs(transp, xdr_argument, &argument)) {
		svcerr_decode(transp);
		return;
	}
	result = (*local) (uid, &argument);
	if (!svc_sendreply(transp, xdr_result, (char *) result)) {
		xlog(L_NOTICE, "unable to send reply.\n");
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, &argument))
		xlog(L_FATAL, "unable to free arguments\n");
}
