/*
 * ypupdated.c	Main service routines for minimal ypupdated implementation
 *		for auth_des (only does publickey.byname updates).
 *
 * The publickey updater accepts the following requests:
 *
 *  *	from root@nismaster:
 *			May create or change any user or host key.
 *			If there is no host key for the NIS master host yet,
 *			only requests for that key are accepted.
 *  *	from root@otherhost:
 *			May change its own host key, and may create or
 *			change user key.
 *  *	from user@anyhost:
 *			May change his/her own key
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <rpc/rpc.h> 
#include <rpc/pmap_clnt.h> 
#include <rpc/key_prot.h> 
#include <rpcsvc/ypupdate_prot.h>
#include <rpcsvc/ypclnt.h>
#include <fcntl.h> 
#include <unistd.h> 
#include <string.h> 
#include <stdio.h>
#include <stdlib.h>

#include "rpcsec.h"
#include "xlog.h"


#define caller_name(rqstp) \
	inet_ntoa(svc_getcaller(rqstp->rq_xprt)->sin_addr)

/*
 * Function prototypes.
 */
static void	ypu_prog_1(struct svc_req *rqstp, register SVCXPRT *transp);
static int	checksubmitter(struct svc_req *rqstp, char *key);
static int	ypupdate_publickey(struct svc_req *rqstp, int *resp,
					char *key, int keylen,
					char *datum, int datumlen);
/* static int	key_getcred(char *netname, key_getcredres *cred); */
static char *	ypopname(int ypop);
static void	usage(void);

static SVCXPRT *udpxprt;
static SVCXPRT *tcpxprt;
static int	allow_nobody = 0;
static char	myhostname[257];

int
main(int argc, char **argv)
{
	int	foreground = 1, pid, c;

	xlog_open("ypupdated");

	while ((c = getopt(argc, argv, "fd:n")) != EOF) {
		switch (c) {
		case 'f':
			foreground = 1;
			break;
		case 'd':
			xlog_sconfig(optarg, 1);
			break;
		case 'n':
			allow_nobody = 1;
			break;
		default:
			usage();
		}
	}

	if (gethostname(myhostname, sizeof(myhostname)) < 0)
		xlog(L_FATAL, "Can't get my own hostname");

	if (!foreground) {
		if ((pid = fork()) < 0) {
			perror("cannot fork");
			return 1;
		}
		if (pid != 0)
			return 0;
		xlog_background();
	}

	pmap_unset(YPU_PROG, YPU_VERS);

	if ((udpxprt = svcudp_create(RPC_ANYSOCK)) == NULL)
		xlog(L_FATAL, "cannot create udp service.");
	if (!svc_register(udpxprt, YPU_PROG, 1, ypu_prog_1, IPPROTO_UDP))
		xlog(L_FATAL, "unable to register (YPU_PROG, YPU_VERS, udp).");

	if ((tcpxprt = svctcp_create(RPC_ANYSOCK, 0, 0)) == NULL)
		xlog(L_FATAL, "cannot create udp service.");
	if (!svc_register(tcpxprt, YPU_PROG, 1, ypu_prog_1, IPPROTO_TCP))
		xlog(L_FATAL, "unable to register (YPU_PROG, YPU_VERS, tcp).");

	svc_run();

	xlog(L_FATAL, "svc_run returned");
	return 0;
}

static void
usage(void)
{
	fprintf(stderr, "usage: rpc.ypupdated [-d fac] [-fn]\n");
	exit(1);
}

static void
ypu_prog_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
	struct ypupdate_args	args;
	xdrproc_t		xdr_arguments;
	int			retval, reply;

	/* NULL procedure doesn't require authentication */
	if (rqstp->rq_proc == 0) {
		svc_sendreply(transp, (xdrproc_t) xdr_void, NULL);
		return;
	}

	/* Make sure that the caller authenticated him/herself with
	 * AUTH_DES.
	 */
	if (rqstp->rq_cred.oa_flavor != AUTH_DES) {
		xlog(L_WARNING, "%s call from %s with weak authentication",
				ypopname(rqstp->rq_proc), caller_name(rqstp));
		svcerr_weakauth(rqstp->rq_xprt);
		return;
	}

	switch (rqstp->rq_proc) {
	case YPU_DELETE:
		xdr_arguments = (xdrproc_t) xdr_ypdelete_args;
		break;
	case YPU_CHANGE:
	case YPU_INSERT:
	case YPU_STORE:
		xdr_arguments = (xdrproc_t) xdr_ypupdate_args;
		break;
	default:
		svcerr_noproc(transp);
		return;
	}

	/* ypupdate_args contain ypdelete_args */
	memset(&args, 0, sizeof(args));
	if (!svc_getargs(transp, xdr_arguments, (void *) &args)) {
		svcerr_decode(transp);
		return;
	}

	reply = 1;
	if (strcmp(args.mapname, PUBLICKEY_MAP) == 0) {
		reply = ypupdate_publickey(rqstp, &retval,
				args.key.yp_buf_val, args.key.yp_buf_len,
				args.datum.yp_buf_val, args.datum.yp_buf_len);
	} else {
		/* Allow only updates to the publickey map for now. */
		xlog(L_WARNING, "client %s tried to change %s entry!\n",
				caller_name(rqstp), args.mapname);
		retval = YPERR_MAP;
	}

	if (reply) {
		if (!svc_sendreply(transp, (xdrproc_t) xdr_u_int, (void*) &retval))
			svcerr_systemerr(transp);
	}

	if (!svc_freeargs(transp, xdr_arguments, (void *) &args))
		xlog(L_FATAL, "unable to free arguments");

	return;
}

/*
 * Check if a host netname is the local hostname or matches the
 * caller's IP address.
 */
static int
check_host_netname(char *hname, struct in_addr *addr)
{
	struct hostent	*hp;
	char		**ap;

	if ((hp = gethostbyname(hname)) == NULL)
		return 0;

	for (ap = hp->h_addr_list; *ap; ap++)
		if (addr->s_addr == ((struct in_addr *)ap)->s_addr)
			return 1;

	return 0;
}

/*
 * Check whether the supplied netname is a user or host name.
 */
static char *
getnetid(char *netname, char *domain)
{
	int	nlen, slen;
	char	*netid;

	slen = strlen(netname);
	if (yp_match(domain, NETID_MAP, netname, slen, &netid, &nlen))
		return NULL;

	netid[nlen] = '\0';
	return netid;
}

/*
 * Check that the submitter is the user requesting the update, or
 * the local root user.
 */
#define REJECT(why) {	\
	xlog(L_WARNING, "rejected %s(%s) request from %s (%s:%d): " why "\n", \
			ypopname(rqstp->rq_proc), keyname, credname, \
			inet_ntoa(sin->sin_addr), ntohs(sin->sin_port)); \
	return 0; \
}

static int
checksubmitter(struct svc_req *rqstp, char *keyname)
{
	struct sockaddr_in	*sin = svc_getcaller(rqstp->rq_xprt);
	struct authdes_cred	*cred;
	char			dummy[HEXKEYBYTES+1];
	char			*credname, *key_at, *cred_at;
	char			*crednetid, *keynetid;

	/* Get the caller's fullname. This field is always filled in
	 * by _svcauth_des, even for requests that transmit only the
	 * nickname. */
	cred = (struct authdes_cred *)(rqstp->rq_clntcred);
	credname = cred->adc_fullname.name;

	/* Check if this is the nobody user */
	if (!getpublickey(credname, dummy)) {
		if (allow_nobody)
			return 1;
		REJECT("nobody credentials");
	}

	if (!(cred_at = strchr(credname, '@')))
		REJECT("malformed credname");
	if (!(key_at = strchr(keyname, '@')))
		REJECT("malformed keyname");

	if (!(crednetid = getnetid(credname, cred_at+1)))
		REJECT("no netid for credential");

	if (!strncmp(crednetid, "0:", 2)) {
		/* This is a host name. root@localhost may do everything,
		 * but other roots are restricted to changing their host
		 * keys and creating/changing user keys.
		 * We may want to have some more sophisticated ACL support
		 * here; e.g. root on other hosts should not be allowed to
		 * create/modify the keys for system users such as bin,
		 * daemon, etc. */
		if (!strcmp(crednetid+2, myhostname))
			return 1;
		free(crednetid);
		if (!(keynetid = getnetid(keyname, key_at+1)))
			REJECT("no netid for key name");
		if (!strncmp(keynetid, "0:", 2) && strcmp(credname, keyname)) {
			free(keynetid);
			REJECT("may not change other host's key");
		}
		free(keynetid);
	} else {
		/* This is a user's netname. Users are only allowed to
		 * change their own key, which must already exist.  */
		free(crednetid);
		if (strcmp(credname, keyname))
			REJECT("user name mismatch");
	}

	return 1;
}
#undef REJECT

/*
 * Perform the actual change to the publickey file.
 * Probably should fork and run ypmake.
 */
static int
ypupdate_publickey(struct svc_req *rqstp, int *resp,
					char *key, int keylen,
					char *datum, int datumlen)
{
	FILE	*wfp, *rfp;
	char	buffer[1024];
	char	keybuf[MAXNETNAMELEN+1], datumbuf[256];
	int	wfd, retries, added, i;
	int	ypop = rqstp->rq_proc;

	/* Don't allow users to sneak in newlines or stuff */
	if (ypop != YPU_DELETE) {
		if (datumlen > sizeof(datumbuf)) {
			xlog(L_WARNING, "%s: huge hexkey (%d bytes) from %s.\n",
				ypopname(rqstp->rq_proc),
				datumlen, caller_name(rqstp));
			xlog(L_WARNING, "        \"%.*s\"\n", datumlen, datum);
			*resp = YPERR_BADARGS;
			return 1;
		}
		for (i = 0; i < datumlen; i++) {
			if (0 <= datum[i] && datum[i] <= 31) {
				xlog(L_WARNING, "%s: illegal character 0x%0x "
					"in publickey update for %s!\n",
					ypopname(rqstp->rq_proc),
					datum[i], key);
				*resp = YPERR_BADARGS;
				return 1;
			}
		}
		memcpy(datumbuf, datum, datumlen);
		datumbuf[datumlen] = '\0';
		datum = datumbuf;
	} else {
		datum = "";
	}

	if (keylen > MAXNETNAMELEN) {
		xlog(L_WARNING, "%s: huge netname (%d bytes) from %s.\n",
				ypopname(rqstp->rq_proc),
				keylen, caller_name(rqstp));
		*resp = YPERR_BADARGS;
		return 1;
	}
	memcpy(keybuf, key, keylen);
	keybuf[keylen] = '\0';
	key = keybuf;

	if (!strcmp(key, "nobody")) {
		xlog(L_WARNING, "refusing %s(nobody) call from %s.\n",
				ypopname(rqstp->rq_proc),
				caller_name(rqstp));
		*resp = YPERR_BADARGS;
		return 1;
	}

	xlog(D_CALL, "%s called %s(%s, %s, %s) from %s.\n",
		((struct authdes_cred *) rqstp->rq_clntcred)->adc_fullname.name,
		ypopname(rqstp->rq_proc), "publickey.byname", key, datum,
		caller_name(rqstp));

	if (!checksubmitter(rqstp, key)) {
		svcerr_weakauth(rqstp->rq_xprt);
		return 0;
	}

	retries = 0;
	while ((wfd = open(PUBLICKEY_LOCK, O_CREAT|O_WRONLY|O_EXCL,0644)) < 0) {
		if (++retries > 5) {
			xlog(L_WARNING, "failed to lock %s after 5 attempts.\n",
					PUBLICKEY_FILE);
			*resp = YPERR_BUSY;
			return 1;
		}
		sleep(1);
	}
	if ((wfp = fdopen(wfd, "w")) == NULL) {
		xlog(L_ERROR, "failed to fdopen(wfd)!\n");
		close(wfd);
		*resp = YPERR_YPERR;
		return 1;
	}
	if ((rfp = fopen(PUBLICKEY_FILE, "r")) == NULL) {
		xlog(L_ERROR, "failed to open %s.\n", PUBLICKEY_FILE);
		fclose(wfp);
		*resp = YPERR_YPERR;
		return 1;
	}

	added = 0;

	while(fgets(buffer, sizeof(buffer), rfp) != NULL) {
		if (!strncmp(buffer, key, keylen) && buffer[keylen] == ' ') {
			switch (ypop) {
			case YPU_INSERT:
				goto error;
			case YPU_DELETE:
				continue;
			default:
				fprintf(wfp, "%s %s\n", key, datum);
				added++;
			}
		} else {
			fputs(buffer, wfp);
		}
	}
	switch (ypop) {
	case YPU_CHANGE:
		if (!added)
			goto error;
		/* fallthru */
	case YPU_DELETE:
		break;
	default:
		if (!added)
			fprintf(wfp, "%s %s\n", key, datum);
	}

	fclose(wfp);
	fclose(rfp);
	if (rename(PUBLICKEY_LOCK, PUBLICKEY_FILE) < 0) {
		xlog(L_ERROR, "couldn't rename %s to %s!\n");
		*resp = YPERR_YPERR;
		return 1;
	}

	/* execve("/bin/sh", "sh", "-c", "cd /var/yp; make"); */

	*resp = 0;
	return 1;

error:
	fclose(wfp);
	fclose(rfp);
	unlink(PUBLICKEY_LOCK);

	*resp = YPERR_KEY;
	return 1;
}

static char *
ypopname(int ypop)
{
	switch (ypop) {
	case YPU_CHANGE:
		return "ypu_change";
	case YPU_INSERT:
		return "ypu_insert";
	case YPU_DELETE:
		return "ypu_delete";
	case YPU_STORE:
		return "ypu_store";
	}

	return "<badproc>";
}

#if 0
/*
 * The key_getcred call to the key server. This is missing from libc.
 */
static int
key_getcred(char *netname, struct unixcred *cred)
{
	struct sockaddr_in	sin;
	struct timeval		to = { 5, 0 };
	CLIENT *		client;
	getcredres		res;
	int			socket, stat;

	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	bzero(sin.sin_zero, sizeof(sin.sin_zero));
	socket = RPC_ANYSOCK;

	if (!(client = clntudp_bufcreate(&sin, KEY_PROG, 1, &to, &socket)))
		return -1;

	res.getcredres_u.cred = cred;
	to.tv_sec = 20;
	stat = clnt_call(client, KEY_GETCRED,
			(xdrproc_t) xdr_netnamestr, &netname,
			(xdrproc_t) xdr_getcredres, &res, &to);
	clnt_destroy(client);
	close(socket);

	return (stat == RPC_SUCCESS && res.status == KEY_SUCCESS)? 0 : -1;
}
#endif
